"""Define a series of endpoints for what locations are supported."""
from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import Any, cast


class Supported:
    """Define an object to supported location API calls."""

    def __init__(self, request: Callable[..., Awaitable]) -> None:
        """Initialize.

        Args:
            request: The request method from the CloudAPI object.
        """
        self._request = request

    async def cities(self, country: str, state: str) -> list[dict[str, Any]]:
        """Return a list of supported cities in a country/state.

        Args:
            state: A state.
            country: A country.

        Returns:
            An API response payload.
        """
        data = await self._request(
            "get", "cities", params={"state": state, "country": country}
        )
        return [d["city"] for d in data["data"]]

    async def countries(self) -> list[dict[str, Any]]:
        """Return an array of all supported countries.

        Returns:
            An API response payload.
        """
        data = await self._request("get", "countries")
        return [d["country"] for d in data["data"]]

    async def states(self, country: str) -> list[dict[str, Any]]:
        """Return a list of supported states in a country.

        Args:
            country: A country.

        Returns:
            An API response payload.
        """
        data = await self._request("get", "states", params={"country": country})
        return [d["state"] for d in data["data"]]

    async def stations(
        self, city: str, state: str, country: str
    ) -> list[dict[str, Any]]:
        """Return a list of supported stations in a city.

        Args:
            city: A city.
            state: A state.
            country: A country.

        Returns:
            An API response payload.
        """
        data = await self._request(
            "get", "stations", params={"city": city, "state": state, "country": country}
        )
        return cast(list[dict[str, Any]], data["data"])



"""Define a client to interact with the AirVisual Cloud API."""
from __future__ import annotations

from json.decoder import JSONDecodeError
from typing import Any, cast

from aiohttp import ClientSession, ClientTimeout

from .air_quality import AirQuality
from .const import LOGGER
from .errors import AirVisualError
from .node import NodeCloudAPI
from .supported import Supported

API_URL_BASE = "https://api.airvisual.com/v2"

DEFAULT_REQUEST_TIMEOUT = 10


class CloudAPIError(AirVisualError):
    """Define an error related to the Cloud API."""

    pass


class InvalidKeyError(CloudAPIError):
    """Define an error when the API key is invalid."""

    pass


class KeyExpiredError(CloudAPIError):
    """Define an error when the API key has expired."""

    pass


class LimitReachedError(CloudAPIError):
    """Define an error when the API limit has been reached."""

    pass


class NoStationError(CloudAPIError):
    """Define an error when there's no station for the data requested."""

    pass


class NotFoundError(CloudAPIError):
    """Define an error for when a location (city or node) cannot be found."""

    pass


class RequestError(CloudAPIError):
    """Define an error related to invalid requests."""

    pass


class UnauthorizedError(CloudAPIError):
    """Define an error related to unauthorized requests."""

    pass


ERROR_CODES: dict[str, type[AirVisualError]] = {
    "api_key_expired": KeyExpiredError,
    "call_limit_reached": LimitReachedError,
    "city_not_found": NotFoundError,
    "feature_not_available": UnauthorizedError,
    "forbidden": UnauthorizedError,
    "incorrect_api_key": InvalidKeyError,
    "no_nearest_station": NoStationError,
    "node not found": NotFoundError,
    "payment required": InvalidKeyError,
    "permission_denied": UnauthorizedError,
    "too_many_requests": LimitReachedError,
}


def raise_on_data_error(data: dict[str, Any]) -> None:
    """Raise an error if the data payload suggests there is one.

    Args:
        data: A response data payload from the API.

    Raises:
        error: An appropriate API error (a subclass of AirVisualError).
    """
    if "data" not in data or data.get("status") == "success":
        return

    try:
        [error] = [
            v for k, v in ERROR_CODES.items() if k in data["data"]["message"].lower()
        ]
    except ValueError:
        error = AirVisualError
    raise error(data)


class CloudAPI:  # pylint: disable=too-few-public-methods
    """Define an object to work with the AirVisual Cloud API."""

    def __init__(self, api_key: str, session: ClientSession | None = None) -> None:
        """Initialize.

        Args:
            api_key: An API key.
            session: An optional aiohttp ClientSession.
        """
        self._api_key = api_key
        self._session: ClientSession | None = session

        self.air_quality = AirQuality(self._request)
        self.node = NodeCloudAPI(self._request)
        self.supported = Supported(self._request)

    async def _request(
        self,
        method: str,
        endpoint: str,
        *,
        base_url: str = API_URL_BASE,
        **kwargs: dict[str, Any],
    ) -> dict[str, Any]:
        """Make a request against the API.

        Args:
            method: An HTTP method.
            endpoint: A relative API endpoint to query.
            base_url: The base API URL to use.
            **kwargs: Extra arguments to send with the API request.

        Returns:
            An API response payload.
        """
        kwargs.setdefault("headers", {})
        kwargs["headers"]["Content-Type"] = "application/json"

        kwargs.setdefault("params", {})
        kwargs["params"]["key"] = self._api_key

        if use_running_session := self._session and not self._session.closed:
            session = self._session
        else:
            session = ClientSession(
                timeout=ClientTimeout(total=DEFAULT_REQUEST_TIMEOUT)
            )

        try:
            async with session.request(
                method, f"{base_url}/{endpoint}", **kwargs
            ) as resp:
                data = await resp.json(content_type=None)
        except JSONDecodeError:
            # The response can't be parsed as JSON, so we'll use its body text
            # in an error:
            response_text = await resp.text()
            data = {"status": "fail", "data": {"message": response_text}}
        finally:
            if not use_running_session:
                await session.close()

        if isinstance(data, str):
            # In some cases, the AirVisual API will return a quoted string in its
            # response body (e.g., "\"node not found\""), which is technically valid
            # JSON. Additionally, AirVisual sets that response's Content-Type header
            # to application/json (#smh). Together, these factors will allow a
            # non-true-JSON  payload to escape the try/except above. So, if we get
            # here, we use the string value (with quotes removed) to raise an error:
            response_text = data.replace('"', "")
            data = {"status": "fail", "data": {"message": response_text}}

        LOGGER.debug("Data received for /%s: %s", endpoint, data)

        raise_on_data_error(data)

        return cast(dict[str, Any], data)



"""Define objects to interact with an AirVisual Node/Pro."""
from __future__ import annotations

import asyncio
import csv
import json
import tempfile
from collections import OrderedDict
from collections.abc import Awaitable, Callable
from functools import partial
from types import TracebackType
from typing import IO, Any, TypeVar, cast, overload

import numpy as np
import smb
from smb.SMBConnection import SMBConnection

from .const import LOGGER
from .errors import AirVisualError

API_URL_BASE = "https://www.airvisual.com/api/v2/node"

DEFAULT_CONNECT_TIMEOUT = 10

SAMBA_HISTORY_PATTERN = "*_AirVisual_values.txt"
SMB_SERVICE = "airvisual"
SMB_USERNAME = "airvisual"

METRIC_AQI_CN = "aqi_cn"
METRIC_AQI_US = "aqi_us"
METRIC_CO2 = "co2"
METRIC_HUMIDITY = "humidity"
METRIC_PM01 = "pm0_1"
METRIC_PM10 = "pm1_0"
METRIC_PM25 = "pm2_5"
METRIC_VOC = "voc"
METRICS_TO_TREND = [
    METRIC_AQI_CN,
    METRIC_AQI_US,
    METRIC_CO2,
    METRIC_HUMIDITY,
    METRIC_PM01,
    METRIC_PM10,
    METRIC_PM25,
    METRIC_VOC,
]

METRIC_MAPPING = {
    "AQI(CN)": METRIC_AQI_CN,
    "AQI(US)": METRIC_AQI_US,
    "CO2(ppm)": METRIC_CO2,
    "Humidity(%RH)": METRIC_HUMIDITY,
    "PM01(ug/m3)": METRIC_PM01,
    "PM10(ug/m3)": METRIC_PM10,
    "PM2_5(ug/m3)": METRIC_PM25,
    "VOC(ppb)": METRIC_VOC,
    "co2_ppm": METRIC_CO2,
    "humidity_RH": METRIC_HUMIDITY,
    "pm01_ugm3": METRIC_PM01,
    "pm10_ugm3": METRIC_PM10,
    "pm25": METRIC_PM25,
    "pm25_AQICN": METRIC_AQI_CN,
    "pm25_AQIUS": METRIC_AQI_US,
    "pm25_ugm3": METRIC_PM25,
    "voc_ppb": METRIC_VOC,
}

TREND_FLAT = "flat"
TREND_INCREASING = "increasing"
TREND_DECREASING = "decreasing"


class NodeProError(AirVisualError):
    """Define an error related to Node/Pro errors."""

    pass


class NodeConnectionError(NodeProError):
    """Define an error connection issues."""

    pass


class InvalidAuthenticationError(NodeProError):
    """Define an error for invalid authentication."""

    pass


def _calculate_trends(
    history: list[OrderedDict], measurements_to_use: int
) -> dict[str, Any]:
    """Calculate the trends of all data points in history data.

    Args:
        history: A list of dict-based measurements.
        measurements_to_use: The number of measurements to include (-1 for all)

    Returns:
        An API response payload.
    """
    if measurements_to_use == -1:
        index_range = np.arange(0, len(history))
    else:
        index_range = np.arange(0, measurements_to_use)

    measured_attributes = set().union(*(d.keys() for d in history))
    metrics_to_trend = measured_attributes.intersection(list(METRICS_TO_TREND))

    trends = {}
    for attribute in metrics_to_trend:
        values = [
            float(value)
            for measurement in history
            for attr, value in measurement.items()
            if attr == attribute
        ]

        if measurements_to_use != -1:
            values = values[-measurements_to_use:]

        index_array = np.array(values)
        linear_fit = np.polyfit(
            index_range,
            index_array,
            1,
        )
        slope = round(linear_fit[0], 2)

        metric = _get_normalized_metric_name(attribute)

        if slope > 0:
            trends[metric] = TREND_INCREASING
        elif slope < 0:
            trends[metric] = TREND_DECREASING
        else:
            trends[metric] = TREND_FLAT

    return trends


def _get_normalized_metric_name(key: str) -> str:
    """Return a normalized string (if it exists) for a metric.

    Args:
        key: A metric name to examine.

    Returns:
        A normalized metric name or the original.
    """
    return METRIC_MAPPING.get(key, key)


class NodeCloudAPI:  # pylint: disable=too-few-public-methods
    """Define an object to work with getting Node info via the Cloud API."""

    def __init__(self, request: Callable[..., Awaitable]) -> None:
        """Initialize.

        Args:
            request: The request method from the CloudAPI object.
        """
        self._request = request

    async def get_by_node_id(self, node_id: str) -> dict[str, Any]:
        """Return cloud API data from a node its ID.

        Args:
            node_id: A Node ID.

        Returns:
            An API response payload.
        """
        data = await self._request("get", node_id, base_url=API_URL_BASE)
        return cast(dict[str, Any], data)


_SambaOperationReturnType = TypeVar(  # pylint: disable=invalid-name
    "_SambaOperationReturnType",
    int,
    list[smb.base.SharedFile],
    list[dict[str, Any]],
    None,
)


class NodeSamba:
    """Define an object to work with getting Node info over Samba."""

    def __init__(self, ip_or_hostname: str, password: str) -> None:
        """Initialize.

        Args:
            ip_or_hostname: An IP address or hostname to a Node.
            password: A Samba password for a Node.
        """
        self._conn = SMBConnection(SMB_USERNAME, password, "pyairvisual", SMB_SERVICE)
        self._connected = False
        self._ip_or_hostname = ip_or_hostname
        self._latest_history = None
        self._loop = asyncio.get_event_loop()

    async def __aenter__(self) -> NodeSamba:
        """Handle the start of a context manager.

        Returns:
            A connected NodeSamba object.
        """
        await self.async_connect()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,  # noqa: F841
        exc_val: BaseException | None,  # noqa: F841
        exc_tb: TracebackType | None,  # noqa: F841
    ) -> None:
        """Handle the end of a context manager.

        Args:
            exc_type: An optional exception if one caused the context manager to close.
            exc_val: The value of the optional exception
            exc_tb: The traceback of the optional exception
        """
        await self.async_disconnect()

    @overload
    async def _execute_samba_operation(
        self, pysmb_func: Callable[..., list[dict[str, Any]]]
    ) -> list[dict[str, Any]]:
        ...

    @overload
    async def _execute_samba_operation(
        self,
        pysmb_func: Callable[[None], bool],
        ip_or_hostname: str,
        *,
        timeout: int = DEFAULT_CONNECT_TIMEOUT,
    ) -> bool:
        ...

    @overload
    async def _execute_samba_operation(
        self,
        pysmb_func: Callable[[str, str, IO[bytes]], None],
        service: str,  # noqa: F841
        filepath: str,
        file_obj: IO[bytes],  # noqa: F841
    ) -> None:
        ...

    @overload
    async def _execute_samba_operation(
        self,
        pysmb_func: Callable[[str, str], list[smb.base.SharedFile]],
        service: str,  # noqa: F841
        filepath: str,
        *,
        pattern: str | None = None,  # noqa: F841
        search: str | None = None,  # noqa: F841
    ) -> list[smb.base.SharedFile]:
        ...

    async def _execute_samba_operation(
        self,
        pysmb_func: Callable[..., _SambaOperationReturnType],
        *args: Any,
        **kwargs: Any,
    ) -> _SambaOperationReturnType:
        """Guard a Samba command with appropriate error handling.

        Args:
            pysmb_func: A pysmb function to run.
            *args: Any args to pass to the pysmb function.
            **kwargs: Any kwargs to pass to the pysmb function.

        Returns:
            Any type supported by pysmb operations.

        Raises:
            InvalidAuthenticationError: Raised on invalid Samba auth.
            NodeConnectionError: Raised on any Samba connection-related error.
            NodeProError: Raised on any unknown error.
        """
        func_with_kwargs = partial(pysmb_func, **kwargs)
        try:
            res = await self._loop.run_in_executor(  # type: ignore[func-returns-value]
                None, func_with_kwargs, *args
            )
        except smb.base.NotConnectedError as err:
            raise NodeConnectionError(f"The Pro unit is not connected: {err}") from err
        except smb.base.NotReadyError as err:
            raise InvalidAuthenticationError(
                f"The Pro unit returned an authentication error: {err}"
            ) from err
        except smb.base.SMBTimeout as err:
            raise NodeConnectionError(
                "Timed out while talking to the Pro unit"
            ) from err
        except ConnectionRefusedError as err:
            raise NodeConnectionError(
                "Couldn't find a Pro unit at the provided IP address"
            ) from err
        except Exception as err:  # pylint: disable=broad-except
            raise NodeProError(err) from err

        return res

    async def _async_get_history_files(self) -> list[smb.base.SharedFile]:
        """Return all the history files on a Samba device.

        Returns:
            A list of Samba file references.
        """
        return await self._execute_samba_operation(
            self._conn.listPath,
            SMB_SERVICE,
            "/",
            pattern=SAMBA_HISTORY_PATTERN,
            search=smb.smb_constants.SMB_FILE_ATTRIBUTE_NORMAL,
        )

    async def _async_retrieve_data_from_tempfile(
        self, tmp_file: IO[bytes]
    ) -> list[dict[str, Any]]:
        """Retrieve data from a NamedTemporaryFile.

        Args:
            tmp_file: A reference to a NamedTemporaryFile.

        Returns:
            An API response payload.
        """

        def get_data() -> list[dict[str, Any]]:
            """Get the data.

            Returns:
                An API response payload.
            """
            data = []
            with open(tmp_file.name, encoding="utf-8") as file:
                reader = csv.DictReader(file, delimiter=";")
                for row in reader:
                    data.append(
                        {
                            _get_normalized_metric_name(header): value
                            for header, value in row.items()
                        }
                    )

            LOGGER.debug("Loaded data from file: %s", data)
            return data

        return await self._execute_samba_operation(get_data)

    async def _async_store_filepath_in_tempfile(
        self, filepath: str, tmp_file: IO[bytes]
    ) -> None:
        """Save a file to a NamedTemporaryFile object.

        Args:
            filepath: A filepath on the Node Samba share.
            tmp_file: A reference to a NamedTemporaryFile.
        """
        await self._execute_samba_operation(
            self._conn.retrieveFile, SMB_SERVICE, filepath, tmp_file
        )

    async def async_connect(self, *, timeout: int = DEFAULT_CONNECT_TIMEOUT) -> None:
        """Connect to the Node.

        Args:
            timeout: The number of seconds before timing out the connection attempt.

        Raises:
            InvalidAuthenticationError: Raised when the provided Samba password
                is incorrect.
        """
        if self._connected:
            LOGGER.debug("Already connected!")
            return

        result = await self._execute_samba_operation(
            self._conn.connect, self._ip_or_hostname, timeout=timeout
        )

        if result:
            self._connected = True
        else:
            raise InvalidAuthenticationError("Invalid Samba authentication")

    async def async_disconnect(self) -> None:
        """Disconnect from the Node."""
        if not self._connected:
            LOGGER.debug("Already disconnected!")
            return

        await self._execute_samba_operation(self._conn.close)
        self._connected = False

    async def async_get_history(
        self, *, include_trends: bool = True, measurements_to_use: int = -1
    ) -> dict[str, Any]:
        """Get history data from the device.

        Args:
            include_trends: Whether trend data should be included.
            measurements_to_use: The number of measurements to include (-1 for all)

        Returns:
            An API response payload.

        Raises:
            NodeProError: Raised when no history files are found.
        """
        history_files = await self._async_get_history_files()
        history_files.sort(key=lambda file: file.filename, reverse=True)

        if not history_files:
            raise NodeProError(
                f"No history files found that match {SAMBA_HISTORY_PATTERN}"
            )

        data: dict[str, Any] = {}

        for history_file in history_files:
            tmp_file = (
                tempfile.NamedTemporaryFile()  # pylint: disable=consider-using-with
            )
            await self._async_store_filepath_in_tempfile(
                f"/{history_file.filename}", tmp_file
            )
            tmp_file.seek(0)
            data["measurements"] = await self._async_retrieve_data_from_tempfile(
                tmp_file
            )
            tmp_file.close()

            if include_trends:
                data["trends"] = _calculate_trends(
                    data["measurements"], measurements_to_use
                )

            if data["measurements"]:
                break

        return data

    async def async_get_latest_measurements(self) -> dict[str, Any]:
        """Get the latest measurements from the device.

        Returns:
            An API response payload.
        """
        data = {}

        tmp_file = tempfile.NamedTemporaryFile()
        await self._async_store_filepath_in_tempfile(
            "/latest_config_measurements.json", tmp_file
        )
        tmp_file.seek(0)
        raw = tmp_file.read()
        tmp_file.close()
        data = json.loads(raw.decode())

        LOGGER.debug("Node measurements loaded: %s", data)

        try:
            # Handle a single measurement returned in a list:
            measurements = data["measurements"][0].items()
        except KeyError:
            # Handle a single measurement returned as a standalone dict:
            measurements = data["measurements"].items()

        data["last_measurement_timestamp"] = int(data["date_and_time"]["timestamp"])
        data["measurements"] = {
            _get_normalized_metric_name(pollutant): value
            for pollutant, value in measurements
        }
        data["status"]["sensor_life"] = {
            _get_normalized_metric_name(pollutant): value
            for pollutant, value in data["status"].get("sensor_life", {}).items()
        }

        return data



"""Define an object to retrieve air quality info."""
from __future__ import annotations

from collections.abc import Awaitable, Callable
from typing import Any, cast


class AirQuality:
    """Define an object to manage air quality API calls."""

    def __init__(self, request: Callable[..., Awaitable]) -> None:
        """Initialize.

        Args:
            request: The request method from the CloudAPI object.
        """
        self._request = request

    async def _nearest(
        self,
        kind: str,
        latitude: float | str | None = None,
        longitude: float | str | None = None,
    ) -> dict[str, Any]:
        """Return data from nearest city/station (IP or coordinates).

        Args:
            kind: "city" or "station".
            latitude: A latitude.
            longitude: A longitude.

        Returns:
            An API response payload.
        """
        if latitude and longitude:
            params = {"lat": str(latitude), "lon": str(longitude)}
        else:
            params = {}

        data = await self._request("get", f"nearest_{kind}", params=params)
        return cast(dict[str, Any], data["data"])

    async def city(self, city: str, state: str, country: str) -> dict[str, Any]:
        """Return data for the specified city.

        Args:
            city: A city.
            state: A state.
            country: A country.

        Returns:
            An API response payload.
        """
        data = await self._request(
            "get", "city", params={"city": city, "state": state, "country": country}
        )
        return cast(dict[str, Any], data["data"])

    async def nearest_city(
        self,
        latitude: float | str | None = None,
        longitude: float | str | None = None,
    ) -> dict[str, Any]:
        """Return data from nearest city (IP or coordinates).

        Args:
            latitude: A latitude.
            longitude: A longitude.

        Returns:
            An API response payload.
        """
        return await self._nearest("city", latitude, longitude)

    async def nearest_station(
        self,
        latitude: float | str | None = None,
        longitude: float | str | None = None,
    ) -> dict[str, Any]:
        """Return data from nearest station (IP or coordinates).

        Args:
            latitude: A latitude.
            longitude: A longitude.

        Returns:
            An API response payload.
        """
        return await self._nearest("station", latitude, longitude)

    async def ranking(self) -> list[dict[str, Any]]:
        """Return a sorted array of selected major cities in the world.

        Returns:
            An API response payload.
        """
        data = await self._request("get", "city_ranking")
        return cast(list[dict[str, Any]], data["data"])

    async def station(
        self, station: str, city: str, state: str, country: str
    ) -> dict[str, Any]:
        """Return data for the specified station.

        Args:
            station: A station name.
            city: A city.
            state: A state.
            country: A country.

        Returns:
            An API response payload.
        """
        data = await self._request(
            "get",
            "station",
            params={
                "station": station,
                "city": city,
                "state": state,
                "country": country,
            },
        )
        return cast(dict[str, Any], data["data"])
